#!/usr/bin/env python
#
#     Name:          malmon
#     Version:       0.1b1
#     Author:        ShadowX
#     Contact:       georgi.kolev AT gmail.com
#     Last Updated:  21.01.2011
#
#  ABOUT:
#   This daemon will monitor file change activity in directory
#   (and sub dirs) and scan (md5 sum check, hex signature check)
#   files and try to match them agains a database of known malware.
#
#  DEPENDS ON:
#     * pyinotify module [ http://pyinotify.sourceforge.net/ ]
#     * inotify kernel support [ CONFIG_INOTIFY_USER=y ]
#
#
#   Copyright 2011 [ShadowX] Georgi Kolev <georgi.kolev at gmail.com>
#
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#   MA 02110-1301, USA.
#
#
#  Changelog:
#   v0.1b1
#     * Its not possible to monitor multiple dirs at once.
#     * Now using our own db with md5 sums and signatures
#       ! Please mail me any script that malmon didn't detect !
#     * Forking LMD's md5 and hex sig files, and switching to
#       our own... and changing the version style :]
#     * Fixed the update procedure.
#     * New md5 sums and hex signatures added
#     * Now supporting file size limit below 1mb.
#     * Blacklist is working again. (no reg expr for now)
#
#   v0.1a3
#     * Forgot to add fork to demonize :P
#     * Fixed install script.
#     * FIxed a broken path.
#     * Fixed forgoten variable in os.path.exists() :p
#     I push those changes at the last minute, and ...Sorry :/
#
#   v0.1a2
#     * Added md5sum and hex signature autoupdate at start
#     * Added option in the config file to choose between
#       working in daemon or normal mode. [daemon - default]
#
#   v0.1a - Fisrt initial release.
#     * Getting md5sums, hex sigs and ideas from LMD
#         http://www.rfxn.com/projects/linux-malware-detect/
#         Really great project! I'm just doing what they are...
#         and I'm using ideas and db's from thir project so all
#         creditz goes to them! (rfxn.com) Great work guys!
#     * Added daemon mode with inotify to monitor file system activity
#     * Added md5sum check
#     * Added hex signature based check
#     * Added exclude_list (you can enter files/dirs to exlude)
#     * Added black_list (list of restricted filesnames)
#     * Switching to Queue put relay new/modified files into one place
#         and the analayzer thread/fAnal/ will pull new stuff FIFO style
#         We do this to lower the load 'cosed by the daemon.
#
#  TODO:
#   * Add mimetype checks
#   * Add word-match-spam-filter like checks
#   * Add walk options, to check dirs without going into daemon mode
#   * Add support to monitor multiple directories at the same time.
#   * Adding multiple threads to handle analyzation of files
#       This will be helpful for servers with intensive changes
#   * Add warn/paranoid mode.
#       Warn mode will only add to log if malware is found
#       Paranoid mode will move the file to quarantine folder

confFile = '/etc/malmon/malmon.conf'

import os, sys, string
import threading, Queue
import hashlib, binascii, urllib
import logging, pyinotify, shutil
from ConfigParser import ConfigParser

from pyinotify import WatchManager, Notifier, ThreadedNotifier, \
		ProcessEvent, IN_MOVED_TO, IN_CLOSE_WRITE, IN_CREATE
mask = IN_CLOSE_WRITE | IN_MOVED_TO | IN_CREATE


class fAnal(threading.Thread):
	"""
	    File analayzer thread.
	    This thread will get (from the Queue)
	    all new/modified files and analyze them.

	    parm @logger     - logger object
	    parm @Q          - Queue (we get new tasks from here)
	    parm @quarantine - Where to move malware files
	    parm @backlist   - List of forbien filenames
	    parm @maxsize    - Max file size (check only files below this size)
	    parm @cachedir   - The deamon cache directory
	"""
	def __init__(self, logger, Q, quarantine, blacklist, maxsize, cachedir='/var/cache/malmon'):
		threading.Thread.__init__(self)
		self.maxsize = (float(maxsize) * (1024*1024))
		self.blacklist = blacklist
		self.cachedir = cachedir
		self.quar = quarantine
		self.logger = logger
		self.alive = False
		self.removed = []
		self.md5dict = {}
		self.hexdict = {}
		self.Q = Q

	def run(self):
		self.alive = True
		# First we load the md5 and hex files
		self.load_md5hex()
		self.shitload()

		while self.alive:
			# Get new files
			newfile = self.Q.get()
			self.logger.debug('Checking %s' % newfile)
			
			# First blacklist checks.
			filename = newfile.split('/')[-1]
			if filename in self.shitlist:
				self.logger.info('[shitlist] Moving %s' % newfile)
				self.move_file(newfile)
				continue
			
			# Look only files smaller then 1Mb!
			try:
				if os.path.getsize(newfile) < self.maxsize:
					filecontent = self.read_file(newfile)
					if not filecontent:
						self.logger.debug('Empty file? %s' % newfile)
						continue

					# md5 check
					md5r = self.check_md5(filecontent)
					if md5r:
						self.logger.warning('[md5] %s is %s. Moving!' % (newfile, md5r.rstrip()))
						self.move_file(newfile)
						self.removed.append(newfile)
						continue
				
					# hex signature check
					hexr = self.check_hex(filecontent)
					if hexr:
						self.logger.warning('[hex] %s is %s. Moving!' % (newfile, hexr.rstrip()))
						self.move_file(newfile)
						self.removed.append(newfile)
						continue
			except OSError:
				if not newfile in self.removed:
					self.logger.debug(
						'OSError rised while checking file size.Maybe file is missing %s ?' % newfile)
			except:
				self.logger.critical('Unknown error while analyzing file %s' % newfile)

	def shitload(self):
		""" This funciton will read the
		    blacklist file and put it into a list
		"""
		try:
			shitf = open(self.shitlist, 'r')
			self.shitlist = shitf.read().split('\n')
			shitf.close()
		except:
			self.shitlist = []
		return


	def move_file(self, newfile):
		""" Move files to quaranty dir
		    parm @newfile - source file
		"""
		try:
			dstfile = os.path.join(self.quar, newfile.split('/')[-1])
			print newfile, dstfile
			shutil.move(newfile, dstfile)
		except IOError:
			self.logger.error('IOError while moving %s' % newfile)
		except:
			self.logger.error('Unknown error while moving %s' % newfile)
		return

	def read_file(self, newfile):
		""" This function just reads the file """
		try:
			newfl = open(newfile, 'r')
			newfl_cont = newfl.read()
			newfl.close()
		except:
			self.logger.warning('Error while opening/reading %s' % newfile)
			return False
		else:
			return newfl_cont

	def check_hex(self, filecontent):
		"""
		    This function matches the file
		    for known hex signatures of malware

		    parm @filecontent - string with file content
		"""
		# Get file contetnt in hex
		hexcontent = str(binascii.hexlify(filecontent))
		# Try to match hex paterns
		for hex_patern in self.hexdict:
			if hex_patern in hexcontent:
				return self.hexdict[hex_patern]
		return False

	def check_md5(self, filecontent):
		"""
		    This function checks the md5sum of the
		    new/modified file and compares it
		    to the md5sums in md5.dat file

		    @filecontent - string with file content
		    @returns True if we match something
		"""
		md5hash_sum = hashlib.md5(filecontent).hexdigest()
		if md5hash_sum in self.md5dict:
			return self.md5dict[md5hash_sum]
		else:
			return False

	def load_md5hex(self):
		""" This functions loads the the
			md5 sum and hex data files """
		md5file = os.path.join(self.cachedir, 'md5.dat')
		hexfile = os.path.join(self.cachedir, 'hex.dat')
		# Load md5 sums
		if os.path.exists(md5file):
			infile = open(md5file, 'r')
			n = 0
			for infile_line in infile.readlines():
				if len(infile_line) > 1:
					self.md5dict[infile_line.split(':')[0]] = infile_line.split(':')[1]
					n += 1
			infile.close()
			self.logger.info('[md5] %s rules added' % (n))
			del n
		else:
			self.logger.warning('%s not found.' % md5file)
		# Load hex signatures
		if os.path.exists(hexfile):
			infile = open(hexfile, 'r')
			n = 0
			for infile_line in infile.readlines():
				if len(infile_line) > 1:
					self.hexdict[infile_line.split(':')[0]] = infile_line.split(':')[1]
					n += 1
			infile.close()
			self.logger.info('[hex] %s rules added.' % (n))
			del n
		else:
			self.logger.warning('%s not found.' % hexfile)

		return


	def finish(self):
		self.alive = False

class EventHandler(pyinotify.ProcessEvent):
	"""
	    This class handles create/modify events
	    and puts event.pathname into queue so
	    the other thread can analyze the file
	"""
	def __init__(self, event):
		if not event.dir:
			Q.put(event.pathname)


def update_definitions(logger, cachedir,
	url='http://darknetx.eu/malmon/'):
	"""
	    This function will download and update
	    the md5 sum and hex signature files.
	    
	    parm @logger   - logger object
	    parm @cachedir - the daemon cache dir
	    parm @newver   - the new version id
	    parm @url      - url -> where md5/hex files can be found
	    
	    Again :) We are using files from the LMD project.
	    All creditz go to them!
	"""
	# URLs
	md5sumurl = '%s/md5.dat' % (url)
	hexsigurl = '%s/hex.dat' % (url)
	verdaturl = '%s/version' % (url)
	# Full pathnames
	md5path = '%s/md5.dat' % (cachedir)
	hexpath = '%s/hex.dat' % (cachedir)
	verpath = '%s/version' % (cachedir)
	
	try:
		webFile = urllib.urlopen(verdaturl)
		localfile = open(verpath, 'w')
		localfile.write(str(webFile.read()).strip())
		webFile.close()
		localfile.close()
	except IOError:
		logger.error('IOError rised while saving maldet.sigs.ver to cache dir')
	try:
		webFile = urllib.urlopen(md5sumurl)
		localfile = open(md5path, 'w')
		localfile.write(webFile.read())
		webFile.close()
		localfile.close()
	except IOError:
		logger.error('IOError rised while saving md5.dat to cache dir')
	try:
		webFile = urllib.urlopen(hexsigurl)
		localfile = open(hexpath, 'w')
		localfile.write(webFile.read())
		webFile.close()
		localfile.close()
	except IOError:
		logger.error('IOError rised while saving hex.dat to cache dir')
	
	return


def chk_def_ver(logger, cachedir,
		url='http://darknetx.eu/malmon/version'):
	"""
	    This function will compare the version
	    of the md5sum and hex files and update
	    them if needed.
	    
	    parm @logger   - logger object
	    parm @cachedir - the daemon cache dir
	    parm @url      - url to definitions version file
	"""
	verfilepath = os.path.join(cachedir, 'version')
	# If we dont find file -> update!
	if not os.path.exists(verfilepath):
		logger.info('No definition verion file found. Downloading new ones...')
		update_definitions(logger, cachedir)
		return
	# Else -> compare versions
	try:
		verf = open(verfilepath, 'r')
		curver = verf.read()
		verf.close()
	except:
		logger.error('Error while opening %s' % verfilepath)
	else:
		versionfile = urllib.urlopen(url)
		remotever = versionfile.read().strip()
		if not str(curver).strip() == remotever:
			logger.info(
				'Local version: %s Remote version: %s. Starting update...' % (curver.strip(),
												remotever))
			update_definitions(logger, cachedir)
		else:
			logger.info('No updates found.')
	return

def forkme(pidfile, cachedir):
	""" double fork """
	try:
		pid = os.fork() # fork first child
	except OSError, e:
		raise Exception, "%s [%d]" % (e.strerror, e.errno)

	if pid == 0:
		os.setsid()
		try:
			pid = os.fork() # fork second child
		except OSError, e:
			raise Exception, "%s [%d]" % (e.strerror, e.errno)

		if pid == 0:
			os.chdir(cachedir) # change current dir to /tmp
			os.umask(0)	 # umask for files created by the daemon
		else:
			os._exit(0)	 # Exit parent (the first child) of the second child.
	else:
		os._exit(0)	 # Exit parent of the first child.

	# Iterate through and close all file descriptors.
	try:
		maxfd = os.sysconf("SC_OPEN_MAX")
	except (AttributeError, ValueError):
		maxfd = 1024

	#for fd in range(0, maxfd):
	#    try: os.close(fd)
	#    except OSError: pass # ERROR, fd wasn't open to begin with (ignored)

	#sys.stdin.close()
	#sys.stdout = NullDevice()
	#sys.stderr = NullDevice()

	mypid = str(os.getpid()) # get current pid

	# Check if program is already running
	try:
		pfl = file(pidfile, 'r')
		thepid = int(pfl.read().strip())
		pfl.close()
	except IOError:
		thepid = False

	# Check if another copy is running
	if thepid:
		if int(thepid) != int(mypid):
			logger.warning('Another copy of the program is running.\n\t\t\t\t  Sending SIGTERM signal to pid: %s' % thepid)
			try:
				os.kill(thepid, SIGTERM); time.sleep(0.1)
			except OSError, err:
				try:
					os.remove(pidfile)
				except:
					logger.error('Unable to remove pidfile')
	return




# Read config file
config = ConfigParser()
config.read(confFile)
# Global options
watch_dir = config.get("global", "watch_dir")
cachedir  = config.get("global", "cache_dir")
pidfile   = config.get("global", "pidfile")
demonize  = config.get("global", "demonize")
# Scan options
quarantine  = config.get("scan", "quarantine")
excludeFile = config.get("scan", "exclude_file")
blacklist   = config.get("scan", "blacklist")
maxfilesize = config.get("scan", "maxsize")
# Logging options
logfile  = config.get("logging", "logfile")
loglevel = config.get("logging", "level")

# pharse watch_dir
watch_dir = watch_dir.split(';')

# loglevel sanity check
if loglevel in ['DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL']:
	loglevel = eval('logging.' + loglevel)
else:
	print 'Using logging default level [INFO]'
	loglevel = logging.INFO
# Daemon flag sanity check
if not str(demonize).isdigit():
	print 'Unrecodnized value in config file: demonize = %s' % demonize
	print 'The program will run in default (daemon) mode.'
	demonize = 1
else:
    demonize = int(demonize)

# Setup logging
logger = logging.getLogger()
logger.setLevel(loglevel)
if demonize:
	handler = logging.FileHandler(logfile)
else:
	handler = logging.StreamHandler()
handler.setLevel(loglevel)
formatter = logging.Formatter(
	'[%(asctime)s]-[%(levelname)s] %(message)s')
handler.setFormatter(formatter)
logger.addHandler(handler)


# Queue object
Q = Queue.Queue()

# daemon mode ?
if demonize:
	forkme(pidfile, cachedir)

# Check/Update definitions
chk_def_ver(logger, cachedir)

# Analayzer thread
fAnal(logger, Q, quarantine, blacklist,
	maxfilesize, cachedir).start()

# Create watch manager object
wm = WatchManager()
# add handler class
notifier = Notifier(wm, EventHandler)
# Exclude patterns from file
excl = pyinotify.ExcludeFilter(excludeFile)
# rec      - recursivly watch directory and sub-directories
# auto_add - auto add new directories
wdd = wm.add_watch(watch_dir, mask, rec=True,
		auto_add=True, exclude_filter=excl)
# OxFF we go :}
notifier.loop(daemonize=False, force_kill=True)
